/*
 * morpheus - main.c
 *
 * this program tries helping to understand the amorphous depuration
 * and recrystallization of SiCx while ion implanation. hopefully the program 
 * will simulate the stabilization of the selforganizing structure in the
 * observed behaviour.
 *
 * refs: 
 *  - J. K. N. Lindner. Habilationsschrift, Universitaet Augsburg.
 *  - Maik Haeberlen. Diplomarbeit, Universitaet Augsburg.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* important defines */
#include "defines.h"

/* function prototypes */
#include "random.h"
#include "display.h"

/* global variables */
u32 sum_z_cells;
int random_fd; /* /dev/urandom file descriptor */

int usage()
{
 puts("usage:");
 puts("-h: help");
 puts("-a <value> \t slope of nuclear energy loss (default 1)");
 puts("-c <value> \t nuclear enery loss at depths 0 (default 0)");
 puts("-x <value> \t # x cells (default 50)");
 puts("-y <value> \t # y cells (default 50)");
 puts("-z <value> \t # z cells (default 100)");
 puts("-s <value> \t # steps to calculate (default 5000)");
 puts("-X <value> \t display area intercept point x (default # x celss / 2)");
 puts("-Y <value> \t display area intercept point y (default # y cells / 2)");
 puts("-Z <value> \t display area intercept point z (default # z cells / 2)");
 puts("-d <value> \t refresh every <value> loops (default 100)");
 puts("-r <value> \t pressure range from amorphous SiCx (default 2)");
 puts("-f <value> \t faktor for pressure from amorphous SiCx (default 1)");
 return -23;
}

int make_amorph(u32 *cell)
{
 *cell|=AMORPH;
 return 23;
}

int make_cryst(u32 *cell)
{
 *cell&=~AMORPH;
 return 23;
}

/* look at cell ... */
int process_cell(void *cell_p,u32 x,u32 y,u32 z,u32 x_max,u32 y_max,u32 z_max,int range,int faktor)
{
 /* is amorph? */
 u32 *cell;
 int i,j,k;
 float count;
 cell=(u32 *)(cell_p+x+y*(x_max-1)+z*(x_max-1)*(y_max-1));
 count=0;
 for(i=-range;i<=range;i++)
  for(j=-range;j<=range;j++)
   for(k=-range;k<=range;k++)
    if(!(i==0 && j==0 && k==0) && (x>=range && x<x_max-range) \
        && (y>=range && y<y_max-range) && (z>=range && z<z_max-range))
     if(*(u32 *)(cell_p+(x+i)+(y+j)*(x_max-1)+(z+k)*(x_max-1)*(y_max-1))&AMORPH)
      // total+=range*range-
      count+=0.01*faktor*1.0/(i*i+j*j+k*k);
 /* count is propability of cell getting amorph */
 /* for now assume 1 - ~ is propability of cell getting cryst if allready amorphous */    
 if(*cell&AMORPH)
 {
  if((1-count)*URAND_2_MAX<=rand_get(URAND_2_MAX))
  {  
   printfd("cell getting cryst. ...(%f)\n",(1-count));
   make_cryst(cell);
  }
 } else
 {
  if(count*URAND_2_MAX<=rand_get(URAND_2_MAX))
  {
   printfd("cell getting amorph ...(%f)\n",count);
   make_amorph(cell);
  }
 }
 return 23;
}

int main(int argc,char **argv) 
{
 u32 x_cell,y_cell,z_cell; /* amount of segments */
 u32 x,y,z; /* cells */
 int i; /* for counting */
 int slope_nel,start_nel; /* nuclear energy loss: slope, constant */
 int a_p_range,a_p_faktor; /* pressure range and faktor from amorphous sic */
 int steps; /* # steps */
 void *cell_p;
 struct __display display;
 u32 display_x,display_y,display_z; /* intercept point of diplayed areas */
 u32 display_refresh_rate; /* refresh rate for display */

 /* default values */
 x_cell=DEFAULT_X_SEG;
 y_cell=DEFAULT_Y_SEG;
 z_cell=DEFAULT_Z_SEG;
 slope_nel=DEFAULT_SLOPE_NEL;
 start_nel=DEFAULT_START_NEL;
 a_p_range=DEFAULT_A_P_RANGE;
 a_p_faktor=DEFAULT_A_P_FAKTOR;
 steps=DEFAULT_STEPS;
 display_x=x_cell/2;
 display_y=y_cell/2;
 display_z=z_cell/2;
 display_refresh_rate=DEFAULT_DISPLAY_REF_RATE;
 
 /* parse command args */
 for(i=1;i<argc;i++)
 {
  if(argv[i][0]=='-')
  {
   switch(argv[i][1])
   {
    case 'h':
     usage();
     return 23;
     break;
    case 'a':
     slope_nel=atoi(argv[++i]);
     break;
    case 'c':
     start_nel=atoi(argv[++i]);
     break;
    case 'x':
     x_cell=atoi(argv[++i]);
     break;
    case 'y':
     y_cell=atoi(argv[++i]);
     break;
    case 'z':
     z_cell=atoi(argv[++i]);
     break;
    case 's':
     steps=atoi(argv[++i]);
     break;
    case 'X':
     display_x=atoi(argv[++i]);
     break;
    case 'Y':
     display_y=atoi(argv[++i]);
     break;
    case 'Z':
     display_z=atoi(argv[++i]);
     break;
    case 'd':
     display_refresh_rate=atoi(argv[++i]);
     break;
    case 'r':
     a_p_range=atoi(argv[++i]);
     break;
    case 'f':
     a_p_faktor=atoi(argv[++i]);
     break;
    default:
     usage();
     return -23;
   }
  } else usage();
 }

 /* open random fd */
 if((random_fd=open("/dev/urandom",O_RDONLY))<0)
 {
  puts("cannot open /dev/urandom\n");
  return -23;
 }

 /* calculate sum_z_cells one time! */
 sum_z_cells=0;
 for(i=1;i<=z_cell;i++) sum_z_cells+=(start_nel+i*slope_nel);
 printfd("debug: sum z cells -> %d\n",sum_z_cells);


 /* testing ... */

 /* allocate cells */
 printf("malloc will free %d bytes now ...\n",x_cell*y_cell*z_cell*sizeof(u32));
 if((cell_p=malloc(x_cell*y_cell*z_cell*sizeof(u32)))==NULL)
 {
  puts("failed allocating memory for cells\n");
  return -23;
 }
 memset(cell_p,0,x_cell*y_cell*z_cell*sizeof(u32));

 /* init display */
 display_init(x_cell,y_cell,z_cell,&display,cell_p,&argc,argv);

 /* main routine */
 for(i=0;i<steps;i++)
 {
  x=rand_get(x_cell);
  y=rand_get(y_cell);
  z=rand_get_lgp(slope_nel,start_nel);

  /* todo */
  // distrib_c_conc(cell_p);

  // process_cell((u32 *)(cell_p+x+y*(x_cell-1)+z*(x_cell-1)*(y_cell-1)));
  process_cell(cell_p,x,y,z,x_cell,y_cell,z_cell,a_p_range,a_p_faktor);

  /* display stuff */
  if((i%display_refresh_rate)==0) 
   display_draw(&display,display_x,display_y,display_z);
 }
  
 /* display again and quit when button hit */
 display_draw(&display,display_x,display_y,display_z);
 puts("hit button to quit ...");
 getchar();

 display_release(&display);

 return 23;
}
